//-----------------------------------------------------------------------------
//
//  Stylophone Studio 5.c
//
//  Stylophone Studio 5
//  A USB MIDI controllable reproduction of the original 1968 Stylophone
//
//  Copyright (c) 2011 Simon Inns
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Web:    http://www.waitingforfriday.com
//  Email:  simon.inns@gmail.com
//
//-----------------------------------------------------------------------------

// Global includes
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/interrupt.h>
#include <stdbool.h>

// Local includes
#include "Descriptors.h"
#include "Stylophone Studio 5.h"
#include "leds.h"
#include "stylophone.h"
#include "buttons.h"
#include "midiCommands.h"
#include "noteStealing.h"
#include "keyboard.h"

// LUFA USB Stack includes
#include "LUFA/Version.h"
#include "LUFA/Drivers/USB/USB.h"

// Globals for hold LED states
uint16_t midiInLedHold;
uint16_t midiInLedState;

uint16_t midiOutLedHold;
uint16_t midiOutLedState;

// Global to hold current pitch bend value
uint16_t currentPitchBendValue;

// Main program loop
int main(void)
{
	setupHardware();
	sei();

	// Initialise state for MIDI indication LED holding
	midiInLedHold = 0;
	midiInLedState = 0;
	midiOutLedHold = 0;
	midiOutLedState = 0;

	while(1)
	{
		// Perform MIDI related tasks
		midiTasks();
		
		// Poll the buttons
		buttonsPoll();
		
		// Check if the 'tune Stylophone' button has been pressed
		if (buttonReadTest(BUTTONS_TUNE_ID) == 0)
		{
			// Process change of tune switch
			if (buttonStatus(BUTTONS_TUNE_ID) == 1)
			{
				// Note: When tuning the Stylophone the USB interface
				// is disabled as tuning requires accurate timing and
				// we can't afford to keep polling the USB tasks during
				// the process.
				
				// Disconnect the USB
				USB_Disable();
				ledOff(LED_VUSB);
				
				// Tune the Stylophone
				stylophoneTune();
				
				// Reconnect the USB
				USB_Init();
				ledOn(LED_VUSB);
			}
		}
		
		// Process midi indication LEDs
		processMidiIndication();
		
		// Poll the keyboard
		keyboardPoll();
		
		// Poll the LUFA USB tasks
		USB_USBTask();
	}
}

// Process MIDI indication LEDs
void processMidiIndication(void)
{
	// Process the MIDI indicator LED holding
	if (midiInLedState == 1)
	{
		if (midiInLedHold == 0) ledOn(LED_MIDIIN);
			
		midiInLedHold++;
			
		if (midiInLedHold > 300)
		{
			midiInLedHold = 0;
			midiInLedState = 0;
			ledOff(LED_MIDIIN);
		}
	}
	
	if (midiOutLedState == 1)
	{
		if (midiOutLedHold == 0) ledOn(LED_MIDIOUT);
			
		midiOutLedHold++;
			
		if (midiOutLedHold > 300)
		{
			midiOutLedHold = 0;
			midiOutLedState = 0;
			ledOff(LED_MIDIOUT);
		}
	}
}

// Set up and configure the hardware
void setupHardware(void)
{
	// Override watchdog enabled fuse
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	// Disable clock division (overrides CPUDIV8 fuse)
	clock_prescale_set(clock_div_1);
	
	// Initialise the hardware and software sub-functions
	ledInitialise();		// Initialise the indicator LEDs
	buttonsInitialise();	// Initialise the hardware buttons
	stylophoneInitialise(); // Initialise the Stylophone sound module
	keyboardInitialise();	// Initialise the keyboard
	initNoteList();			// Initialise the note-stealing function
	
	// Initialise the LUFA stack
	USB_Init();

	// Set the current pitch bend value
	currentPitchBendValue = 8192; // Centered
 }

// This event handler is called when a USB connect event is detected
void EVENT_USB_Device_Connect(void)
{
	// Indicate USB enumerating
}

// This event handler is called when a USB disconnect event is detected
void EVENT_USB_Device_Disconnect(void)
{
	// Indicate USB not ready
	ledOff(LED_VUSB);
}

// This event handler is called after a connect event when the host sets the configuration for the device
void EVENT_USB_Device_ConfigurationChanged(void)
{
	bool configSuccess = true;

	// Setup MIDI Data Endpoints
	configSuccess &= Endpoint_ConfigureEndpoint(MIDI_STREAM_IN_EPNUM, EP_TYPE_BULK, ENDPOINT_DIR_IN,
	                                            MIDI_STREAM_EPSIZE, ENDPOINT_BANK_SINGLE);
	configSuccess &= Endpoint_ConfigureEndpoint(MIDI_STREAM_OUT_EPNUM, EP_TYPE_BULK, ENDPOINT_DIR_OUT,
	                                            MIDI_STREAM_EPSIZE, ENDPOINT_BANK_SINGLE);

	// Indicate endpoint configuration success or failure
	if (configSuccess) ledOn(LED_VUSB); else ledOff(LED_VUSB);
}

// Process MIDI messages in and out of the device
void midiTasks(void)
{
	uint8_t midiCommand = 0;
	uint8_t midiData2 = 0;
	uint8_t midiData3 = 0;
	uint8_t midiChannel = 0;

	// Here we process MIDI out (to host) events.  This is processed whether or not the USB
	// is available as we want the Stylophone to work as a stand-alone instrument.

	// Test to see if we have processed the vibrato switch already
	if (buttonReadTest(BUTTONS_VIBRATO_ID) == 0)
	{
		// Process change of vibrato switch
		if (buttonStatus(BUTTONS_VIBRATO_ID) == 1)
		{
			stylophoneVibrato(VIBRATO_ON);
				
			midiCommand = MIDI_CIN_CONTROL_CHANGE;
			midiData2 = 1;
			midiData3 = 127;
		}				
		else
		{
			stylophoneVibrato(VIBRATO_OFF);
				
			midiCommand = MIDI_CIN_CONTROL_CHANGE;
			midiData2 = 1;
			midiData3 = 0;
		}				
	}
		
	// Test to see if we have processed the change tuning switch already
	if (buttonReadTest(BUTTONS_CHANGE_TUNING_ID) == 0)
	{
		// Process the change tuning button
		if (buttonStatus(BUTTONS_CHANGE_TUNING_ID) == 1)
		{
			// Cancel all notes
			allNotesOff();
				
			midiCommand = MIDI_CIN_PROGRAM_CHANGE;
				
			if (stylophoneTuningStatus() == TUNING_BASS)
			{
				stylophoneTuning(TUNING_STANDARD);
				midiData3 = 1;
			}					
			else if (stylophoneTuningStatus() == TUNING_STANDARD)
			{
				stylophoneTuning(TUNING_TREBLE);
				midiData3 = 2;
			}					
			else
			{
				stylophoneTuning(TUNING_BASS);
				midiData3 = 0;
			}					
		}
	}
		
	// Process keyboard
	uint8_t keyNumber = keyboardGetKeyOffEvents();
	if (keyNumber != 0)
	{
		// Process key off event from keyboard
		midiCommand = MIDI_CIN_NOTE_OFF;
			
		// To determine the MIDI note number we have to take into account
		// the current tuning mode of the Stylophone
		// Bass tuning: (A1 to E3) - midi note 33 to midi note 52
		// Standard tuning: (A2 to E4) - midi note 45 to midi note 64
		// Treble tuning: (A3 to E5) - midi note 57 to midi note 76
				
		switch(stylophoneTuningStatus())
		{
			case TUNING_BASS:
				keyNumber += 32;
				break;
						
			case TUNING_STANDARD:
				keyNumber += 44;
				break;
						
			case TUNING_TREBLE:
				keyNumber += 56;
				break;
		}						
				
		// Set the MIDI key number
		midiData2 = keyNumber;
				
		// Set the velocity (minimum)
		midiData3 = 0;
			
		// Stop the note on the Stylophone
		noteOff(midiData2, currentPitchBendValue);
	}
	else
	{
		// Check for key on events
		keyNumber = keyboardGetKeyOnEvents();
			
		if (keyNumber != 0)
		{
			// Process key on event from keyboard
			midiCommand = MIDI_CIN_NOTE_ON;
				
			// To determine the MIDI note number we have to take into account
			// the current tuning mode of the Stylophone
			// Bass tuning: (A1 to E3) - midi note 33 to midi note 52
			// Standard tuning: (A2 to E4) - midi note 45 to midi note 64
			// Treble tuning: (A3 to E5) - midi note 57 to midi note 76
				
			switch(stylophoneTuningStatus())
			{
				case TUNING_BASS:
					keyNumber += 32;
					break;
						
				case TUNING_STANDARD:
					keyNumber += 44;
					break;
						
				case TUNING_TREBLE:
					keyNumber += 56;
					break;
			}						
				
			// Set the MIDI key number
			midiData2 = keyNumber;
				
			// Set the velocity
			midiData3 = 100;
				
			// Play the note on the Stylophone
			noteOn(midiData2, currentPitchBendValue);
		}
	}							

	// Check if a MIDI command is to be sent (and that the USB is ready to send it)
	if (midiCommand && USB_DeviceState == DEVICE_STATE_Configured)
	{
		Endpoint_SelectEndpoint(MIDI_STREAM_IN_EPNUM);

		if (Endpoint_IsINReady())
		{
			// Indicate that a MIDI out message is being sent
			midiOutLedHold = 0;
			midiOutLedState = 1;
		
			MIDI_EventPacket_t MIDIEvent = (MIDI_EventPacket_t)
				{
					.CableNumber = 0,
					.Command     = midiCommand,

					.Data1       = (midiCommand << 4) | midiChannel,
					.Data2       = midiData2,
					.Data3       = midiData3,
				};

			// Write the MIDI event packet to the endpoint
			Endpoint_Write_Stream_LE(&MIDIEvent, sizeof(MIDIEvent), NULL);

			// Send the data in the endpoint to the host
			Endpoint_ClearIN();
		}
	}
	
	// Process MIDI messages coming in from host ------------------------------------------------------------

	// Check that the USB device is connected and configured
	if (USB_DeviceState == DEVICE_STATE_Configured)
	{
		// Select the MIDI OUT stream
		Endpoint_SelectEndpoint(MIDI_STREAM_OUT_EPNUM);

		// Check if a MIDI command has been received
		if (Endpoint_IsOUTReceived())
		{
			MIDI_EventPacket_t MIDIEvent;

			// Indicate that a MIDI in message has been received
			midiInLedHold = 0;
			midiInLedState = 1;

			// Read the MIDI event packet from the endpoint
			Endpoint_Read_Stream_LE(&MIDIEvent, sizeof(MIDIEvent), NULL);

			// Process MIDI commands:
			switch(MIDIEvent.Command)
			{
				case MIDI_CIN_NOTE_ON:
					// Midi note on command
					
					// Range check the MIDI note to make sure it is valid
					//
					// Firstly we must test to see if the note can be played by the Stylophone.  Supported notes
					// (for each tuning mode) are as follows:
					//
					// Bass tuning: (A1 to E3) - midi note 33 to midi note 52
					// Standard tuning: (A2 to E4) - midi note 45 to midi note 64
					// Treble tuning: (A3 to E5) - midi note 57 to midi note 76
				
					// Is tuning set to Bass?
					if (stylophoneTuningStatus() == TUNING_BASS)
					{
						if (MIDIEvent.Data2 < 33 || MIDIEvent.Data2 > 52)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}
				
					// Is tuning set to Standard?
					if (stylophoneTuningStatus() == TUNING_STANDARD)
					{
						if (MIDIEvent.Data2 < 45 || MIDIEvent.Data2 > 64)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}	
				
					// Is tuning set to Treble?
					if (stylophoneTuningStatus() == TUNING_TREBLE)
					{
						if (MIDIEvent.Data2 < 57 || MIDIEvent.Data2 > 76)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}
				
					// If the note is valid, play it
					if (MIDIEvent.Data2 != 0)
					{
						// Note: Some midi controllers send a note on command with zero
						// velocity instead of note off... We have to check for this
						if (MIDIEvent.Data3 != 0)
							noteOn(MIDIEvent.Data2, currentPitchBendValue);					
						else
							noteOff(MIDIEvent.Data2, currentPitchBendValue); // Bad note off command RX'd
					}
					break; // MIDI_CIN_NOTE_ON
				
				case MIDI_CIN_NOTE_OFF:
					// Range check the MIDI note to make sure it is valid
			
					// Is tuning set to Bass?
					if (stylophoneTuningStatus() == TUNING_BASS)
					{
						if (MIDIEvent.Data2 < 33 || MIDIEvent.Data2 > 52)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}
				
					// Is tuning set to Standard?
					if (stylophoneTuningStatus() == TUNING_STANDARD)
					{
						if (MIDIEvent.Data2 < 45 || MIDIEvent.Data2 > 64)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}	
				
					// Is tuning set to Treble?
					if (stylophoneTuningStatus() == TUNING_TREBLE)
					{
						if (MIDIEvent.Data2 < 57 || MIDIEvent.Data2 > 76)
							MIDIEvent.Data2 = 0; // Set note number to zero
					}
				
					// If the note is valid, turn it off
					if (MIDIEvent.Data2 != 0)
					{
							noteOff(MIDIEvent.Data2, currentPitchBendValue);
					}
					break; // MIDI_CIN_NOTE_OFF
				
				case MIDI_CIN_CONTROL_CHANGE:
	        		// Midi controller change event
	        	
	        		// Note: We support control 1 (modulation) for vibrato control
	        		if (MIDIEvent.Data2 == 1)
	        		{
		        		// Modulation control (the Stylophone only supports on and off)
		        		if (MIDIEvent.Data3 == 0) stylophoneVibrato(VIBRATO_OFF);
		        			else stylophoneVibrato(VIBRATO_ON);
					}
					break;
		        
				case MIDI_CIN_PROGRAM_CHANGE:
		    		// Midi program change event (select tuning mode - 1 = bass, 2 = standard, 3 = treble)
		    	
					// Cancel all playing notes
					allNotesOff();
				
					// Set the Stylophone's current tuning
					switch(MIDIEvent.Data2)
					{
						case 0:	stylophoneTuning(TUNING_BASS);
								break;
							
						case 1:	stylophoneTuning(TUNING_STANDARD);
								break;
							
						case 2:	stylophoneTuning(TUNING_TREBLE);
								break;
					}
		    		break;
	        	
				case MIDI_CIN_PITCH_BEND_CHANGE:
	        		// Pitch-bend change
				
	        		// Note: midiData3 contains the 7 MSBs of the pitch-bend value
	        		//       midiData2 contains the 7 LSB
	        		currentPitchBendValue = MIDIEvent.Data2 + ((MIDIEvent.Data3 << 7) & 0xFF80);
	        	
	        		// Adjust the pitch of the current note
	        		adjustNotePitch(currentPitchBendValue);
							
	        		break;
			
				default:
					break;

 			} // end switch()

			// If the endpoint is now empty, clear the bank
			if (!(Endpoint_BytesInEndpoint()))
			{
				// Clear the endpoint ready for new packet
				Endpoint_ClearOUT();
			}		
		} // endif
	} // endif		
}

