//-----------------------------------------------------------------------------
//
//  buttons.c
//
//  Stylophone Studio 5
//  A USB MIDI controllable reproduction of the original 1968 Stylophone
//
//  Copyright (c) 2011 Simon Inns
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Web:    http://www.waitingforfriday.com
//  Email:  simon.inns@gmail.com
//
//-----------------------------------------------------------------------------

// Global includes
#include <avr/io.h>

// Local includes
#include "buttons.h"

// Globals for storing button states
uint8_t buttonState[3];
uint8_t debounceCounter[3];
uint8_t buttonRead[3];

// Initialise the buttons
void buttonsInitialise(void)
{
	// Set up port directions and pull-ups
	DDRB  &= ~BUTTONS_PORTB_BUTTONS;
	PORTB |=  BUTTONS_PORTB_BUTTONS;
	
	DDRE  &= ~BUTTONS_PORTE_BUTTONS;
	PORTE |=  BUTTONS_PORTE_BUTTONS;
	
	// Initialise button states
	for (uint8_t buttonNumber = 0; buttonNumber < 3; buttonNumber++)
	{
		buttonState[buttonNumber] = 0;
		debounceCounter[buttonNumber] = 0;
		buttonRead[buttonNumber] = 0;
	}
}

// Poll the buttons and perform debounce
void buttonsPoll(void)
{
	// Vibrato button
	if (!(PINE & BUTTONS_VIBRATO))
	{
		// Physical button is active (debounce on)
		if (buttonState[BUTTONS_VIBRATO_ID] == 0)
		{
			debounceCounter[BUTTONS_VIBRATO_ID]++;
			if (debounceCounter[BUTTONS_VIBRATO_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_VIBRATO_ID] = 1;
				buttonRead[BUTTONS_VIBRATO_ID] = 0;
				debounceCounter[BUTTONS_VIBRATO_ID] = 0;
			}
		}
	} 
	else
	{
		// Physical button is inactive (debounce off)
		if (buttonState[BUTTONS_VIBRATO_ID] == 1)
		{
			debounceCounter[BUTTONS_VIBRATO_ID]++;
			if (debounceCounter[BUTTONS_VIBRATO_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_VIBRATO_ID] = 0;
				buttonRead[BUTTONS_VIBRATO_ID] = 0;
				debounceCounter[BUTTONS_VIBRATO_ID] = 0;
			}
		}
	}
	
	// Change tuning button
	if (!(PINB & BUTTONS_CHANGE_TUNING))
	{
		// Physical button is pressed (debounce off)
		if (buttonState[BUTTONS_CHANGE_TUNING_ID] == 0)
		{
			debounceCounter[BUTTONS_CHANGE_TUNING_ID]++;
			if (debounceCounter[BUTTONS_CHANGE_TUNING_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_CHANGE_TUNING_ID] = 1;
				buttonRead[BUTTONS_CHANGE_TUNING_ID] = 0;
				debounceCounter[BUTTONS_CHANGE_TUNING_ID] = 0;
			}
		}
	} 
	else
	{
		// Physical button is released (debounce on)
		if (buttonState[BUTTONS_CHANGE_TUNING_ID] == 1)
		{
			debounceCounter[BUTTONS_CHANGE_TUNING_ID]++;
			if (debounceCounter[BUTTONS_CHANGE_TUNING_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_CHANGE_TUNING_ID] = 0;
				buttonRead[BUTTONS_CHANGE_TUNING_ID] = 0;
				debounceCounter[BUTTONS_CHANGE_TUNING_ID] = 0;
			}
		}
	}
	
	// Tune button
	if (!(PINB & BUTTONS_TUNE))
	{
		// Physical button is pressed (debounce on)
		if (buttonState[BUTTONS_TUNE_ID] == 0)
		{
			debounceCounter[BUTTONS_TUNE_ID]++;
			if (debounceCounter[BUTTONS_TUNE_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_TUNE_ID] = 1;
				buttonRead[BUTTONS_TUNE_ID] = 0;
				debounceCounter[BUTTONS_TUNE_ID] = 0;
			}
		}
	} 
	else
	{
		// Physical button is released (debounce off)
		if (buttonState[BUTTONS_TUNE_ID] == 1)
		{
			debounceCounter[BUTTONS_TUNE_ID]++;
			if (debounceCounter[BUTTONS_TUNE_ID] > BUTTONS_DEBOUNCE_AMOUNT)
			{
				buttonState[BUTTONS_TUNE_ID] = 0;
				buttonRead[BUTTONS_TUNE_ID] = 0;
				debounceCounter[BUTTONS_TUNE_ID] = 0;
			}
		}
	}
}

// Return the status of a button (0 = off, 1 = on)
uint8_t buttonStatus(uint8_t button)
{
	uint8_t status = 0;
	
	if (button == BUTTONS_VIBRATO_ID) status = buttonState[BUTTONS_VIBRATO_ID];
	if (button == BUTTONS_CHANGE_TUNING_ID) status = buttonState[BUTTONS_CHANGE_TUNING_ID];
	if (button == BUTTONS_TUNE_ID) status = buttonState[BUTTONS_TUNE_ID];
	
	return status;
}

// Check to see if the present button state has already been read
uint8_t buttonReadTest(uint8_t button)
{
	uint8_t status = 0;
	
	if (button == BUTTONS_VIBRATO_ID)
	{
		status = buttonRead[BUTTONS_VIBRATO_ID];
		buttonRead[BUTTONS_VIBRATO_ID] = 1;
	}
			
	if (button == BUTTONS_CHANGE_TUNING_ID)
	{
		status = buttonRead[BUTTONS_CHANGE_TUNING_ID];
		buttonRead[BUTTONS_CHANGE_TUNING_ID] = 1;
	}		
	
	if (button == BUTTONS_TUNE_ID)
	{
		status = buttonRead[BUTTONS_TUNE_ID];
		buttonRead[BUTTONS_TUNE_ID] = 1;
	}
	
	return status;
}