//-----------------------------------------------------------------------------
//
//  keyboard.c
//
//  Stylophone Studio 5
//  A USB MIDI controllable reproduction of the original 1968 Stylophone
//
//  Copyright (c) 2011 Simon Inns
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Web:    http://www.waitingforfriday.com
//  Email:  simon.inns@gmail.com
//
//-----------------------------------------------------------------------------

// Global includes
#include <avr/io.h>

// Local includes
#include "keyboard.h"

// Globals for storing current key on and key off events
uint8_t keyboardState[21];
uint8_t keyboardDebounceCounter[21];
uint8_t keyboardRead[21];

// Initialise the Stylophone keyboard
void keyboardInitialise(void)
{
	uint8_t keyNumber = 0;
	
	// Configure ADMUX for  AVcc Vref - Right justified
	ADMUX = ((1 << REFS0));
	
	// Configure ADCSRA for ADC Enabled - /128 prescaler
	ADCSRA = ((1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0));
	
	// Configure ADCSRB for high-speed ADC mode
	ADCSRB = (1 << ADHSM);
	
	// Initialise keyboard state
	for (keyNumber = 1; keyNumber < 21; keyNumber++)
	{
		keyboardState[keyNumber] = 0;
		keyboardDebounceCounter[keyNumber] = 0;
		keyboardRead[keyNumber] = KEY_READ;
	}		
}

// Get the current key number (or 0 for no key pressed)
uint8_t keyboardPerformADC(void)
{
	uint16_t adcResult0 = 0;
	uint16_t adcResult1 = 0;

	uint8_t keyNumber = 0;
	
	// Clear the previous result
	ADCH = 0;
	ADCL = 0;
	
	// Select ADC channel 0
	ADMUX &= ~((1 << MUX5) | (1 << MUX4) | (1 << MUX3) | (1 << MUX2) | (1 << MUX1) | (1 << MUX0));
	
	// Set ADSC to start an ADC conversion
	ADCSRA |= (1 << ADSC);

	// Wait for the ADC conversion to complete
	while(ADCSRA & (1 << ADSC));
	
	// Store the result
	adcResult0 = ADCL + (ADCH << 8);
	
	// Clear the previous result
	ADCH = 0;
	ADCL = 0;
	
	// Select ADC channel 1
	ADMUX |= (1 << MUX0);
	
	// Set ADSC to start an ADC conversion
	ADCSRA |= (1 << ADSC);

	// Wait for the ADC conversion to complete
	while(ADCSRA & (1 << ADSC));
	
	// Store the result
	adcResult1 = ADCL + (ADCH << 8);

	// Lower half of keyboard pressed
	if (adcResult0 >= ADC_POSL10 && adcResult0 <= ADC_POSH10) keyNumber = 10;
	else if (adcResult0 >= ADC_POSL9 && adcResult0 <= ADC_POSH9) keyNumber = 9;
	else if (adcResult0 >= ADC_POSL8 && adcResult0 <= ADC_POSH8) keyNumber = 8;
	else if (adcResult0 >= ADC_POSL7 && adcResult0 <= ADC_POSH7) keyNumber = 7;
	else if (adcResult0 >= ADC_POSL6 && adcResult0 <= ADC_POSH6) keyNumber = 6;
	else if (adcResult0 >= ADC_POSL5 && adcResult0 <= ADC_POSH5) keyNumber = 5;
	else if (adcResult0 >= ADC_POSL4 && adcResult0 <= ADC_POSH4) keyNumber = 4;
	else if (adcResult0 >= ADC_POSL3 && adcResult0 <= ADC_POSH3) keyNumber = 3;
	else if (adcResult0 >= ADC_POSL2 && adcResult0 <= ADC_POSH2) keyNumber = 2;
	else if (adcResult0 >= ADC_POSL1 && adcResult0 <= ADC_POSH1) keyNumber = 1;

	// Upper half of keyboard pressed
	if (adcResult1 >= ADC_POSL10 && adcResult1 <= ADC_POSH10) keyNumber = 20;
	else if (adcResult1 >= ADC_POSL9 && adcResult1 <= ADC_POSH9) keyNumber = 19;
	else if (adcResult1 >= ADC_POSL8 && adcResult1 <= ADC_POSH8) keyNumber = 18;
	else if (adcResult1 >= ADC_POSL7 && adcResult1 <= ADC_POSH7) keyNumber = 17;
	else if (adcResult1 >= ADC_POSL6 && adcResult1 <= ADC_POSH6) keyNumber = 16;
	else if (adcResult1 >= ADC_POSL5 && adcResult1 <= ADC_POSH5) keyNumber = 15;
	else if (adcResult1 >= ADC_POSL4 && adcResult1 <= ADC_POSH4) keyNumber = 14;
	else if (adcResult1 >= ADC_POSL3 && adcResult1 <= ADC_POSH3) keyNumber = 13;
	else if (adcResult1 >= ADC_POSL2 && adcResult1 <= ADC_POSH2) keyNumber = 12;
	else if (adcResult1 >= ADC_POSL1 && adcResult1 <= ADC_POSH1) keyNumber = 11;
	
	return keyNumber;
}

// Poll the keyboard (and perform debounce)
void keyboardPoll(void)
{
	uint8_t keyNumber = 0;
	uint8_t keyPressed = 0;
	
	// Read the keyboard to determine if a key is pressed
	keyPressed = keyboardPerformADC();
	
	for (keyNumber = 1; keyNumber < 21; keyNumber++)
	{
		// Perform debounce
		if (keyNumber == keyPressed)
		{
			// Physical button is pressed (debounce on)
			if (keyboardState[keyNumber] == 0)
			{
				keyboardDebounceCounter[keyNumber]++;
				if (keyboardDebounceCounter[keyNumber] > KEYBOARD_DEBOUNCE_AMOUNT)
				{
					keyboardState[keyNumber] = 1;
					keyboardRead[keyNumber] = KEY_NOT_READ;
					keyboardDebounceCounter[keyNumber] = 0;
				}
			}
		} 
		else
		{
			// Physical button is released (debounce off)
			if (keyboardState[keyNumber] == 1)
			{
				keyboardDebounceCounter[keyNumber]++;
				if (keyboardDebounceCounter[keyNumber] > KEYBOARD_DEBOUNCE_AMOUNT)
				{
					keyboardState[keyNumber] = 0;
					keyboardRead[keyNumber] = KEY_NOT_READ;
					keyboardDebounceCounter[keyNumber] = 0;
				}
			}
		}
	}
}

// Get key-off events
uint8_t keyboardGetKeyOffEvents(void)
{
	uint8_t keyNumber = 0;
	uint8_t returnKey = 0;
	
	// Search for an unread key-off event
	for (keyNumber = 1; keyNumber < 21; keyNumber++)
	{
		if (keyboardState[keyNumber] == 0 && keyboardRead[keyNumber] == KEY_NOT_READ)
		{
			returnKey = keyNumber;
			keyboardRead[keyNumber] = KEY_READ;
			break;
		}
	}		
		
	return returnKey;
}

// Get key-on events
uint8_t keyboardGetKeyOnEvents(void)
{
	uint8_t keyNumber = 0;
	uint8_t returnKey = 0;
	
	// Search for an unread key-off event
	for (keyNumber = 1; keyNumber < 21; keyNumber++)
	{
		if (keyboardState[keyNumber] == 1 && keyboardRead[keyNumber] == KEY_NOT_READ)
		{
			returnKey = keyNumber;
			keyboardRead[keyNumber] = KEY_READ;
			break;
		}
	}		
		
	return returnKey;
}