//-----------------------------------------------------------------------------
//
//  noteStealing.c
//
//  Stylophone Studio 5
//  A USB MIDI controllable reproduction of the original 1968 Stylophone
//
//  Copyright (c) 2011 Simon Inns
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Web:    http://www.waitingforfriday.com
//  Email:  simon.inns@gmail.com
//
//-----------------------------------------------------------------------------

#ifndef NOTESTEALING_C
#define NOTESTEALING_C

// Global includes
#include <avr/io.h>

// Local includes
#include "notestealing.h"
#include "stylophone.h"

// Globals
// Globals for linked list
uint8_t sllArray[MAXIMUM_NOTES][2];
uint8_t sllStartPointer;
uint8_t sllNumberOfEntries;

// Note: Since the Stylophone is a monophonic instrument (it can only play
// one note at a time) we have to use a 'note stealing' algorithm to handle
// polyphonic note requests from the midi controller.
//
// We use a basic algorithm which always plays the latest note received and then
// reverts to the previous note(s) as note-off commands are received.  This allows
// the musician to hold a note and then tap another to legato between the notes.
//
// The algorithm supports a maximum of 10 held notes at any one time.

// Turn a new note on
void noteOn(uint8_t note, uint16_t pitchBendValue)
{
	// Add the note
	addNoteToList(note);
	
	// Play the current note
	switch(stylophoneTuningStatus())
	{
		case TUNING_BASS:
			// Bass tuning (A1 to E3)
			setDac(getDacValueForNote(getMostRecentNoteFromList() - 33, TUNING_BASS));
			break;
					
		case TUNING_STANDARD:
			// Standard tuning (A2 to E4)
			setDac(getDacValueForNote(getMostRecentNoteFromList() - 45, TUNING_STANDARD));
			break;
					
		case TUNING_TREBLE:
			// Treble tuning (A3 to E5)
			setDac(getDacValueForNote(getMostRecentNoteFromList() - 57, TUNING_TREBLE));
			break;
	}
	
	// Check for pitch bend
	if (pitchBendValue != 8192) adjustNotePitch(pitchBendValue);
}

// Turn a note off
void noteOff(uint8_t note, uint16_t pitchBendValue)
{
	// Add the note
	deleteNoteFromList(note);
	
	if (getMostRecentNoteFromList() != SLL_NULL)
	{
		// Play the current note
		switch(stylophoneTuningStatus())
		{
			case TUNING_BASS:
				// Bass tuning (A1 to E3)
				setDac(getDacValueForNote(getMostRecentNoteFromList() - 33, TUNING_BASS));
				break;
					
			case TUNING_STANDARD:
				// Standard tuning (A2 to E4)
				setDac(getDacValueForNote(getMostRecentNoteFromList() - 45, TUNING_STANDARD));
				break;
					
			case TUNING_TREBLE:
				// Treble tuning (A3 to E5)
				setDac(getDacValueForNote(getMostRecentNoteFromList() - 57, TUNING_TREBLE));
				break;
		}
		
		// Check for pitch bend
		if (pitchBendValue != 8192) adjustNotePitch(pitchBendValue);
	}
	else setDac(0);
}

// Reset all notes
void allNotesOff(void)
{
	// Clear the note list
	initNoteList();
	
	// Turn off the DAC
	setDac(0);
}			

// Adjust the pitch of the current note
void adjustNotePitch(uint16_t pitchBendValue)
{
	uint16_t upperDacValue = 0, lowerDacValue = 0;
	uint8_t currentNote = 0;
	uint16_t currentNoteDacValue = 0;
	
	// Only perform pitch bending if there is a current note playing
	currentNote = getMostRecentNoteFromList(); // Note: returns midi note value
	
	if (currentNote != SLL_NULL)
	{
		// Get the DAC value for the upper and lower notes (depending on the tuning)
		switch(stylophoneTuningStatus())
		{
			case TUNING_BASS:
				// Bass tuning (A1 to E3)
				if (currentNote != 52) // Can't pitch bend above E3
					upperDacValue = getDacValueForNote(currentNote - 32, TUNING_BASS);
				else
					upperDacValue = getDacValueForNote(currentNote - 33, TUNING_BASS);
				
				if (currentNote != 33) // Can't pitch bend below A1
					lowerDacValue = getDacValueForNote(currentNote - 34, TUNING_BASS);
				else
					lowerDacValue = getDacValueForNote(currentNote - 33, TUNING_BASS);
					
				break;
					
			case TUNING_STANDARD:
				// Standard tuning (A2 to E4)
				if (currentNote != 64) // Can't pitch bend above E4
					upperDacValue = getDacValueForNote(currentNote - 44, TUNING_STANDARD);
				else
					upperDacValue = getDacValueForNote(currentNote - 45, TUNING_STANDARD);
				
				if (currentNote != 45) // Can't pitch bend below A2
					lowerDacValue = getDacValueForNote(getMostRecentNoteFromList() - 46, TUNING_STANDARD);
				else
					lowerDacValue = getDacValueForNote(currentNote - 45, TUNING_STANDARD);
				break;
					
			case TUNING_TREBLE:
				// Treble tuning (A3 to E5)
				if (currentNote != 76) // Can't pitch bend above E5
					upperDacValue = getDacValueForNote(currentNote - 56, TUNING_TREBLE);
				else
					upperDacValue = getDacValueForNote(currentNote - 57, TUNING_TREBLE);
				
				if (currentNote != 57) // Can't pitch bend below A3
					lowerDacValue = getDacValueForNote(currentNote - 58, TUNING_TREBLE);
				else
					lowerDacValue = getDacValueForNote(currentNote - 57, TUNING_TREBLE);
				break;
		}
		
		// Get the current note's true DAC value
		switch(stylophoneTuningStatus())
		{
			case TUNING_BASS:
				// Bass tuning (A1 to E3)
				currentNoteDacValue = getDacValueForNote(currentNote - 33, TUNING_BASS);
				break;
					
			case TUNING_STANDARD:
				// Standard tuning (A2 to E4)
				currentNoteDacValue = getDacValueForNote(currentNote - 45, TUNING_STANDARD);
				break;
					
			case TUNING_TREBLE:
				// Treble tuning (A3 to E5)
				currentNoteDacValue = getDacValueForNote(currentNote - 57, TUNING_TREBLE);
				break;
		}
		
		// Perform pitch bend
		if (pitchBendValue == 8192)
		{
			// Pitch bend is centered, play the true note
			setDac(currentNoteDacValue);
		}
		else if (pitchBendValue < 8192)
		{
			// Pitch bend down
			pitchBendValue = 8192 - pitchBendValue; // Center bend on zero
			pitchBendValue = pitchBendValue / 100; // Remove some resolution
			
			uint16_t difference = currentNoteDacValue - lowerDacValue;
			difference = (difference * pitchBendValue) / 81;
			setDac(currentNoteDacValue - difference);
		}
		else
		{
			// Pitch bend up
			pitchBendValue = pitchBendValue - 8192; // Center bend on zero
			pitchBendValue = pitchBendValue / 100; // Remove some resolution
			
			uint16_t difference = upperDacValue - currentNoteDacValue;
			difference = (difference * pitchBendValue) / 81;
			setDac(currentNoteDacValue + difference);
		}	
	}
}	

// Initialise the note list
void initNoteList(void)
{
	// Reset the linked list storage array to null
	uint8_t i;
	for (i = 0; i < MAXIMUM_NOTES; i++)
	{
		sllArray[i][SLL_VALUE] = SLL_NULL;
		sllArray[i][SLL_POINTER] = SLL_NULL;
	}
	
	// Reset the linked list start pointer
	sllStartPointer = 0;
	sllNumberOfEntries = 0;
}

// Add a note to the linked list
void addNoteToList(uint8_t noteValue)
{
	// Ensure we do not overflow the stack
	if (sllNumberOfEntries >= MAXIMUM_NOTES) return;
	
	// Make sure the noteValue is not a reserved value
	if (noteValue == SLL_END || noteValue == SLL_NULL) return;
	
	// Make sure the new note value is unique in the list
	//unsigned char targetNoteNumber = 0;
	//while((sllArray[targetNoteNumber][SLL_VALUE] != noteValue)
		//&& (targetNoteNumber < MAXIMUM_NOTES))
		//targetNoteNumber++;
	//if (targetNoteNumber != MAXIMUM_NOTES) return;
	
	// Find the end of the list
	uint8_t endOfListNoteNumber = sllStartPointer;
	
	// We only need to seek the end if there are entries in the list
	if (sllNumberOfEntries != 0)
	{
		while(sllArray[endOfListNoteNumber][SLL_POINTER] != SLL_END)
			endOfListNoteNumber++;
	}
	
	// Seek for a free note in our list
	uint8_t nextFreeNoteNumber = 0;
	while(sllArray[nextFreeNoteNumber][SLL_POINTER] != SLL_NULL)
		nextFreeNoteNumber++;
		
	// Add the new entry
	sllArray[nextFreeNoteNumber][SLL_VALUE] = noteValue;
	sllArray[nextFreeNoteNumber][SLL_POINTER] = SLL_END;
	
	// Point to the new entry
	if (sllNumberOfEntries != 0)
	{
		sllArray[endOfListNoteNumber][SLL_POINTER] = nextFreeNoteNumber;
	}
	
	// Increment the number of entries
	sllNumberOfEntries++;
}

// Delete a note from the linked list
void deleteNoteFromList(uint8_t noteValue)
{
	// Check for an empty list
	if (sllNumberOfEntries == 0) return;
	
	// Find the target note
	uint8_t targetNoteNumber = 0;
	while((sllArray[targetNoteNumber][SLL_VALUE] != noteValue)
		&& (targetNoteNumber < MAXIMUM_NOTES))
		targetNoteNumber++;
		
	// Check if note value was found
	if (targetNoteNumber == MAXIMUM_NOTES) return;
	
	// Does the list contain more than one entry?
	if (sllNumberOfEntries > 1)
	{
		// List contains multiple entries
		
		// Are we deleting the first list note?
		if (targetNoteNumber == sllStartPointer)
		{
			// Target note is the start note
			
			// Set the start pointer to the next note in the list
			sllStartPointer = sllArray[targetNoteNumber][SLL_POINTER];
			
			// Delete the entry
			sllArray[targetNoteNumber][SLL_VALUE] = SLL_NULL;
			sllArray[targetNoteNumber][SLL_POINTER] = SLL_NULL;
			
			// Decrement the list counter
			sllNumberOfEntries--;
		}
		else
		{
			// Deletion target note is not first in the list
			
			// Find the note before the target note
			uint8_t previousNoteNumber = 0;
			while(sllArray[previousNoteNumber][SLL_POINTER] != targetNoteNumber)
				previousNoteNumber++;
				
			// Assign the current note's pointer to the previous note
			sllArray[previousNoteNumber][SLL_POINTER] = 
				sllArray[targetNoteNumber][SLL_POINTER];
			
			// Delete the entry
			sllArray[targetNoteNumber][SLL_VALUE] = SLL_NULL;
			sllArray[targetNoteNumber][SLL_POINTER] = SLL_NULL;
			
			// Decrement the list counter
			sllNumberOfEntries--;
		}
	}
	else
	{
		// List contains only one entry
		
		// Delete the entry
		sllArray[targetNoteNumber][SLL_VALUE] = SLL_NULL;
		sllArray[targetNoteNumber][SLL_POINTER] = SLL_NULL;
		
		// Reset the linked list start pointer
		sllStartPointer = 0;
		
		// Reset the list counter
		sllNumberOfEntries = 0;
	}
}

// Get the most recent note from the linked list
uint8_t getMostRecentNoteFromList(void)
{
	// Check for an empty list
	if (sllNumberOfEntries == 0) return SLL_NULL;
	
	// Seek the last entry in the list
	uint8_t lastNoteNumber = 0;
	while(sllArray[lastNoteNumber][SLL_POINTER] != SLL_END)
		lastNoteNumber++;
		
	// Return the value
	return sllArray[lastNoteNumber][SLL_VALUE];
}

#endif