//-----------------------------------------------------------------------------
//
//  stylophone.c
//
//  Stylophone Studio 5
//  A USB MIDI controllable reproduction of the original 1968 Stylophone
//
//  Copyright (c) 2011 Simon Inns
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Web:    http://www.waitingforfriday.com
//  Email:  simon.inns@gmail.com
//
//-----------------------------------------------------------------------------

// Global includes
#include <avr/io.h>
#include <util/delay.h>
#include <avr/eeprom.h>

// Local includes
#include "stylophone.h"
#include "notevalues.h"
#include "leds.h"

// EEPROM mapping for tuning with default data
uint16_t EEMEM bassTuningFrequenciesEeprom[20];
uint16_t EEMEM standardTuningFrequenciesEeprom[20];
uint16_t EEMEM trebleTuningFrequenciesEeprom[20];

// Tuning data
uint16_t bassTuningFrequencies[20];
uint16_t standardTuningFrequencies[20];
uint16_t trebleTuningFrequencies[20];

// Global for storing tuning status
uint8_t tuningStatus;

// Internal functions ---------------------------------------------------------

// Initialise the 10-bit high-speed PWM function
void initialisePwm(void)
{
	// We are using PC7 for high-speed 10-bit PWM output to the DAC (OC4A)

	// Set OC4A (PC7) output pin direction to output
	DDRC |= (1 << 7);

	// Set the timer PLL to 96Mhz (with clock dividing for USB)
	PLLCSR = (1 << PINDIV) | (1 << PLLE);
	PLLFRQ = (1 << PLLTM0) | (1 << PDIV3) | (1 << PDIV1) | (1 << PLLUSB);

	// Initiate TIMER4 for 10-bit high-speed PWM output
	// Set the Timer/Counter Control Register
	TCCR4A = (1 << COM4A1) | (1 << PWM4A);
	
	// Reset timer 4
	TC4H = 0;
	TCNT4 = 0;
	
	// Set the OCR4A initial duty-cycle to 0%
	TC4H = 0;
	OCR4A = 0;
	
	// Define TOP in OCR4C
	TC4H = 0xFF;
	OCR4C = 0xFF;
	
	// Turn ON the TIMER4 with no prescaler
	TCCR4B |= (1 << CS40);
}

// Set the 16 bit DAC output value
void setDac(uint16_t dacValue)
{
	if (dacValue > 0)
	{
		// Set the MSB duty-cycle
		TC4H = dacValue >> 8;
	
		// Set the LSB duty-cycle
		OCR4A = dacValue;
		
		DDRC |= (DAC_OUT_PIN); // Set pin to output
	}		
	else
	{
		// Set duty cycle to 0%
		TC4H = 0;
		OCR4A = 0;
	}	
}

// Calculate the DAC value for a frequency using maths (see documentation for details
// on how this works)
uint16_t calculateDacValueForFrequency(uint16_t frequency, uint8_t tuningMode)
{
	double dacValueFloat = 0;
	
	switch(tuningMode)
	{
		case TUNING_STANDARD:
			dacValueFloat = 33759 * (double)frequency / 12500;
			break;
			
		case TUNING_BASS:
			dacValueFloat = 33759 * (double)frequency / 6250;
			break;
			
		case TUNING_TREBLE:
			dacValueFloat = 613800 * (double)frequency / 483559;
			break;
	}
	
	frequency = (uint16_t)dacValueFloat;
	
	if (frequency > 1023) frequency = 0;
	
	return frequency;
}

// Measure the current frequency of the oscillator
uint16_t measureFrequency(void)
{
	uint16_t frequency;
	
	// Set up timer1 to act as the pulse counter
	// Timer1 rising edge external clock source
	TCCR1B |= ((1 << CS12) | (1 << CS11) | (1 << CS10));
	
	// Reset timer1 to zero
	TCNT1 = 0;
		
	// Set up timer3 to time 1 second duration for measurement
		
	// Reset timer3 to zero
	TCNT3 = 0;
		
	// Clock has PLL and is running at 96Mhz
	// Start timer with a /256 prescaler
	TCCR3B |= (1 << CS32);
	
	// Allow the frequency signal
	DDRD  &= ~FREQ_CONTROL_PIN;
		
	// Wait 1,000,000 uS (adjusted for code overhead)
	while (TCNT3 < 62350);
		
	// Disallow the frequency signal
	DDRD |= FREQ_CONTROL_PIN;
		
	// Stop timer3
	TCCR3B = 0;
	
	// Read the counter
	frequency = TCNT1;

	return frequency;
}

// Measure the required DAC value for a given frequency
uint16_t measureDacValueForFrequency(uint16_t frequency, uint8_t tuningMode)
{
	uint16_t low = 0;
	uint16_t mid = 0;
	uint16_t high = 0;
	uint16_t test = 0;
	
	// Set the initial range based on the calculated DAC value for the frequency
	low = calculateDacValueForFrequency(frequency, tuningMode) - TUNING_ERROR_RANGE;
	high = calculateDacValueForFrequency(frequency, tuningMode) + TUNING_ERROR_RANGE;
	
	// Search for the correct DAC value
	while (high-low > 1)
	{
		mid = ((high-low)/2)+low;
		setDac(mid);
		
		// Allow the oscillator to settle
		_delay_ms(40);
		
		test = measureFrequency();
		
		if (test == frequency) break;
		if (test > frequency) high = mid;
		if (test < frequency) low = mid;
	}
	
	// Return the correct DAC value for the note
	return mid;	
}

// External functions ---------------------------------------------------------

// Initialise the Stylophone
void stylophoneInitialise(void)
{
	// Set port directions to output for output pins
	//DDRD |= (STANDARD_PIN | BASS_PIN | TREBLE_PIN | VIBRATO_PIN |
	//	TREBLE_ATK_PIN | STANDARD_ATK_PIN);
	
	DDRC |= (STANDARD_GAIN);
	
	DDRD |= (VIBRATO_PIN | STANDARD_PIN | BASS_PIN | TREBLE_PIN |
		TREBLE_ATK_PIN | STANDARD_ATK_PIN);
	
	PORTC &= ~(STANDARD_GAIN);
	
	PORTD &= ~(VIBRATO_PIN | STANDARD_PIN | BASS_PIN | TREBLE_PIN |
		TREBLE_ATK_PIN | STANDARD_ATK_PIN);
	
	// Set frequency measurement pin to input
	DDRD  &= ~FREQ_IN_PIN;
	PORTD &= ~FREQ_IN_PIN;
	
	// Set the control pin to disallow frequency input
	DDRD  |= FREQ_CONTROL_PIN;
	PORTD &= ~FREQ_CONTROL_PIN;
	
	// Set tuning to standard
	stylophoneTuning(TUNING_STANDARD);
	
	// Set vibrato to off
	stylophoneVibrato(VIBRATO_OFF);
	
	// Initialise the PWM output module
	initialisePwm();
	
	// Copy the tuning data from EEPROM to local RAM ready for use
	for (uint8_t note = 0; note < 20; note++)
	{
		bassTuningFrequencies[note] = eeprom_read_word((uint16_t*)&bassTuningFrequenciesEeprom[note]);
		standardTuningFrequencies[note] = eeprom_read_word((uint16_t*)&standardTuningFrequenciesEeprom[note]);
		trebleTuningFrequencies[note] = eeprom_read_word((uint16_t*)&trebleTuningFrequenciesEeprom[note]);
	}
}

// Set the Stylophone's tuning mode
void stylophoneTuning(uint8_t tuningMode)
{
	switch(tuningMode)
	{
		case TUNING_BASS:
			// Set all tuning and attack pins to off
			PORTC &= ~(STANDARD_GAIN);
			PORTD &= ~(STANDARD_PIN | BASS_PIN | TREBLE_PIN | TREBLE_ATK_PIN | STANDARD_ATK_PIN);
			
			// Set tuning to bass
			PORTD |= (STANDARD_PIN | BASS_PIN | STANDARD_ATK_PIN);
			
			// Indicate current tuning
			ledOn(LED_BIRED);
			ledOff(LED_BIGREEN);
			
			tuningStatus = TUNING_BASS;
			break;
			
		case TUNING_STANDARD:
			// Set all pins to off
			PORTC &= ~(STANDARD_GAIN);
			PORTD &= ~(STANDARD_PIN | BASS_PIN | TREBLE_PIN | TREBLE_ATK_PIN | STANDARD_ATK_PIN);
			
			// Set tuning to standard
			PORTC |= (STANDARD_GAIN);
			PORTD |= (STANDARD_PIN | STANDARD_ATK_PIN);
			
			// Indicate current tuning
			ledOn(LED_BIRED);
			ledOn(LED_BIGREEN);
			
			tuningStatus = TUNING_STANDARD;
			break;
			
		case TUNING_TREBLE:
			// Set all pins to off
			PORTC &= ~(STANDARD_GAIN);
			PORTD &= ~(STANDARD_PIN | BASS_PIN | TREBLE_PIN | TREBLE_ATK_PIN | STANDARD_ATK_PIN);
			
			// Set tuning to treble
			PORTC |= (STANDARD_GAIN);
			PORTD |= (TREBLE_PIN | TREBLE_ATK_PIN);
			
			// Indicate current tuning
			ledOff(LED_BIRED);
			ledOn(LED_BIGREEN);
			
			tuningStatus = TUNING_TREBLE;
			break;
	}
}

// Get the tuning status
uint8_t stylophoneTuningStatus(void)
{
	return tuningStatus;
}

// Set the Stylophone's vibrato setting
void stylophoneVibrato(uint8_t vibratoSetting)
{
	if (vibratoSetting == VIBRATO_OFF) PORTD |= VIBRATO_PIN;
	if (vibratoSetting == VIBRATO_ON)  PORTD &= ~VIBRATO_PIN;
}

// Tune the Stylophone
void stylophoneTune(void)
{	
	uint16_t dacValue;
	
	// Get bass tuning DAC values
	stylophoneTuning(TUNING_BASS);
	uint16_t noteNumber = 0;
	for (uint8_t note = 0; note < 20; note++)
	{
		// Get the dacValue for the note and write to EEPROM
		dacValue = measureDacValueForFrequency(noteFrequency[noteNumber], TUNING_BASS);
		eeprom_write_word((uint16_t*)&bassTuningFrequenciesEeprom[note], dacValue);
		bassTuningFrequencies[note] = dacValue;
		noteNumber++;
	}
	
	// Get standard tuning DAC values
	stylophoneTuning(TUNING_STANDARD);
	noteNumber = 12;
	for (uint8_t note = 0; note < 20; note++)
	{
		dacValue = measureDacValueForFrequency(noteFrequency[noteNumber], TUNING_STANDARD);
		eeprom_write_word((uint16_t*)&standardTuningFrequenciesEeprom[note], dacValue);
		standardTuningFrequencies[note] = dacValue;
		noteNumber++;
	}

	// Get treble tuning DAC values
	stylophoneTuning(TUNING_TREBLE);
	noteNumber = 24;
	for (uint8_t note = 0; note < 20; note++)
	{
		dacValue = measureDacValueForFrequency(noteFrequency[noteNumber], TUNING_TREBLE);
		eeprom_write_word((uint16_t*)&trebleTuningFrequenciesEeprom[note], dacValue);
		trebleTuningFrequencies[note] = dacValue;
		noteNumber++;
	}
	
	// Turn the sound off
	setDac(0);
}

// Get DAC value for note from RAM
uint16_t getDacValueForNote(uint8_t note, uint8_t tuning)
{
	uint16_t result = 0;
	
	if (tuning == TUNING_BASS) result = bassTuningFrequencies[note];
	if (tuning == TUNING_STANDARD) result = standardTuningFrequencies[note];
	if (tuning == TUNING_TREBLE) result = trebleTuningFrequencies[note];
	
	return result;
}